@@ -2,7 +2,7 @@
  * Incremental.c - support --incremental.  Part of:
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2006 Neil Brown <neilb@suse.de>
+ * Copyright (C) 2006-2009 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -37,7 +37,8 @@ static void find_reject(int mdfd, struct supertype *st, struct mdinfo *sra,
 			char *array_name);
 
 int Incremental(char *devname, int verbose, int runstop,
-		struct supertype *st, char *homehost, int autof)
+		struct supertype *st, char *homehost, int require_homehost,
+		int autof)
 {
 	/* Add this device to an array, creating the array if necessary
 	 * and starting the array if sensible or - if runstop>0 - if possible.
@@ -48,7 +49,8 @@ int Incremental(char *devname, int verbose, int runstop,
 	 * 2/ Find metadata, reject if none appropriate (check
 	 *       version/name from args)
 	 * 3/ Check if there is a match in mdadm.conf
-	 * 3a/ if not, check for homehost match.  If no match, reject.
+	 * 3a/ if not, check for homehost match.  If no match, assemble as
+	 *    a 'foreign' array.
 	 * 4/ Determine device number.
 	 * - If in mdadm.conf with std name, use that
 	 * - UUID in /var/run/mdadm.map  use that
@@ -56,6 +58,7 @@ int Incremental(char *devname, int verbose, int runstop,
 	 * - Choose a free, high number.
 	 * - Use a partitioned device unless strong suggestion not to.
 	 *         e.g. auto=md
+	 *   Don't choose partitioned for containers.
 	 * 5/ Find out if array already exists
 	 * 5a/ if it does not
 	 * - choose a name, from mdadm.conf or 'name' field in array.
@@ -67,6 +70,7 @@ int Incremental(char *devname, int verbose, int runstop,
 	 * - add the device
 	 * 6/ Make sure /var/run/mdadm.map contains this array.
 	 * 7/ Is there enough devices to possibly start the array?
+	 *     For a container, this means running Incremental_container.
 	 * 7a/ if not, finish with success.
 	 * 7b/ if yes,
 	 * - read all metadata and arrange devices like -A does
@@ -74,20 +78,22 @@ int Incremental(char *devname, int verbose, int runstop,
 	 *   start the array (auto-readonly).
 	 */
 	struct stat stb;
-	struct mdinfo info, info2;
+	struct mdinfo info;
 	struct mddev_ident_s *array_list, *match;
 	char chosen_name[1024];
 	int rv;
-	int devnum;
 	struct map_ent *mp, *map = NULL;
 	int dfd, mdfd;
 	char *avail;
 	int active_disks;
+	int trustworthy = FOREIGN;
+	char *name_to_use;
+	mdu_array_info_t ainf;
+
 	struct createinfo *ci = conf_get_create_info();
-	char *name;
 
 
-	/* 1/ Check if devices is permitted by mdadm.conf */
+	/* 1/ Check if device is permitted by mdadm.conf */
 
 	if (!conf_test_dev(devname)) {
 		if (verbose >= 0)
@@ -137,9 +143,10 @@ int Incremental(char *devname, int verbose, int runstop,
 		close(dfd);
 		return 1;
 	}
-	st->ss->getinfo_super(st, &info);
 	close (dfd);
 
+	memset(&info, 0, sizeof(info));
+	st->ss->getinfo_super(st, &info);
 	/* 3/ Check if there is a match in mdadm.conf */
 
 	array_list = conf_get_ident(NULL);
@@ -148,7 +155,7 @@ int Incremental(char *devname, int verbose, int runstop,
 		if (array_list->uuid_set &&
 		    same_uuid(array_list->uuid, info.uuid, st->ss->swapuuid)
 		    == 0) {
-			if (verbose >= 2)
+			if (verbose >= 2 && array_list->devname)
 				fprintf(stderr, Name
 					": UUID differs from %s.\n",
 					array_list->devname);
@@ -156,7 +163,7 @@ int Incremental(char *devname, int verbose, int runstop,
 		}
 		if (array_list->name[0] &&
 		    strcasecmp(array_list->name, info.name) != 0) {
-			if (verbose >= 2)
+			if (verbose >= 2 && array_list->devname)
 				fprintf(stderr, Name
 					": Name differs from %s.\n",
 					array_list->devname);
@@ -164,7 +171,7 @@ int Incremental(char *devname, int verbose, int runstop,
 		}
 		if (array_list->devices &&
 		    !match_oneof(array_list->devices, devname)) {
-			if (verbose >= 2)
+			if (verbose >= 2 && array_list->devname)
 				fprintf(stderr, Name
 					": Not a listed device for %s.\n",
 					array_list->devname);
@@ -172,7 +179,7 @@ int Incremental(char *devname, int verbose, int runstop,
 		}
 		if (array_list->super_minor != UnSet &&
 		    array_list->super_minor != info.array.md_minor) {
-			if (verbose >= 2)
+			if (verbose >= 2 && array_list->devname)
 				fprintf(stderr, Name
 					": Different super-minor to %s.\n",
 					array_list->devname);
@@ -182,7 +189,7 @@ int Incremental(char *devname, int verbose, int runstop,
 		    !array_list->name[0] &&
 		    !array_list->devices &&
 		    array_list->super_minor == UnSet) {
-			if (verbose  >= 2)
+			if (verbose >= 2 && array_list->devname)
 				fprintf(stderr, Name
 			     ": %s doesn't have any identifying information.\n",
 					array_list->devname);
@@ -191,37 +198,54 @@ int Incremental(char *devname, int verbose, int runstop,
 		/* FIXME, should I check raid_disks and level too?? */
 
 		if (match) {
-			if (verbose >= 0)
-				fprintf(stderr, Name
+			if (verbose >= 0) {
+				if (match->devname && array_list->devname)
+					fprintf(stderr, Name
 		   ": we match both %s and %s - cannot decide which to use.\n",
-					match->devname, array_list->devname);
+						match->devname, array_list->devname);
+				else
+					fprintf(stderr, Name
+						": multiple lines in mdadm.conf match\n");
+			}
 			return 2;
 		}
 		match = array_list;
 	}
 
+	if (match && match->devname
+	    && strcasecmp(match->devname, "<ignore>") == 0) {
+		if (verbose >= 0)
+			fprintf(stderr, Name ": array containing %s is explicitly"
+				" ignored by mdadm.conf\n",
+				devname);
+		return 1;
+	}
+
 	/* 3a/ if not, check for homehost match.  If no match, continue
 	 * but don't trust the 'name' in the array. Thus a 'random' minor
 	 * number will be assigned, and the device name will be based
 	 * on that. */
-	name = info.name;
-	if (!match) {
-		if (homehost == NULL ||
-		    st->ss->match_home(st, homehost) == 0) {
-			if (verbose >= 0)
-				fprintf(stderr, Name
-	      ": not found in mdadm.conf and not identified by homehost.\n");
-			name = NULL;
-		}
+	if (match)
+		trustworthy = LOCAL;
+	else if (st->ss->match_home(st, homehost) == 1)
+		trustworthy = LOCAL;
+	else if (st->ss->match_home(st, "any") == 1)
+		trustworthy = LOCAL_ANY;
+	else
+		trustworthy = FOREIGN;
+
+
+	if (!match && !conf_test_metadata(st->ss->name,
+					  (trustworthy == LOCAL))) {
+		if (verbose >= 1)
+			fprintf(stderr, Name
+				": %s has metadata type %s for which "
+				"auto-assembly is disabled\n",
+				devname, st->ss->name);
+		return 1;
 	}
-	/* 4/ Determine device number. */
-	/* - If in mdadm.conf with std name, get number from name. */
-	/* - UUID in /var/run/mdadm.map  get number from mapping */
-	/* - If name is suggestive, use that. unless in use with */
-	/*           different uuid. */
-	/* - Choose a free, high number. */
-	/* - Use a partitioned device unless strong suggestion not to. */
-	/*         e.g. auto=md */
+	if (trustworthy == LOCAL_ANY)
+		trustworthy = LOCAL;
 
 	/* There are three possible sources for 'autof':  command line,
 	 * ARRAY line in mdadm.conf, or CREATE line in mdadm.conf.
@@ -233,86 +257,72 @@ int Incremental(char *devname, int verbose, int runstop,
 	if (autof == 0)
 		autof = ci->autof;
 
-	if (match && (rv = is_standard(match->devname, &devnum))) {
-		devnum = (rv > 0) ? (-1-devnum) : devnum;
-	} else if ((mp = map_by_uuid(&map, info.uuid)) != NULL)
-		devnum = mp->devnum;
-	else {
-		/* Have to guess a bit. */
-		int use_partitions = 1;
-		char *np, *ep;
-		if ((autof&7) == 3 || (autof&7) == 5)
-			use_partitions = 0;
-		np = name ? strchr(name, ':') : ":NONAME";
-		if (np)
-			np++;
-		else
-			np = name;
-		devnum = strtoul(np, &ep, 10);
-		if (ep > np && *ep == 0) {
-			/* This is a number.  Let check that it is unused. */
-			if (mddev_busy(use_partitions ? (-1-devnum) : devnum))
-				devnum = -1;
-		} else
-			devnum = -1;
-
-		if (devnum < 0) {
-			/* Haven't found anything yet, choose something free */
-			devnum = find_free_devnum(use_partitions);
-
-			if (devnum == NoMdDev) {
-				fprintf(stderr, Name
-					": No spare md devices!!\n");
-				return 2;
-			}
-		} else
-			devnum = use_partitions ? (-1-devnum) : devnum;
+	if (st->ss->container_content && st->loaded_container) {
+		/* This is a pre-built container array, so we do something
+		 * rather different.
+		 */
+		return Incremental_container(st, devname, verbose, runstop,
+					     autof, trustworthy);
 	}
-	mdfd = open_mddev_devnum(match ? match->devname : NULL,
-				 devnum,
-				 name,
-				 chosen_name, autof >> 3);
+
+	name_to_use = info.name;
+	if (name_to_use[0] == 0 &&
+	    info.array.level == LEVEL_CONTAINER &&
+	    trustworthy == LOCAL) {
+		name_to_use = info.text_version;
+		trustworthy = METADATA;
+	}
+	if (name_to_use[0] && trustworthy != LOCAL &&
+	    ! require_homehost &&
+	    conf_name_is_free(name_to_use))
+		trustworthy = LOCAL;
+
+	/* strip "hostname:" prefix from name if we have decided
+	 * to treat it as LOCAL
+	 */
+	if (trustworthy == LOCAL && strchr(name_to_use, ':') != NULL)
+		name_to_use = strchr(name_to_use, ':')+1;
+
+	/* 4/ Check if array exists.
+	 */
+	map_lock(&map);
+	mp = map_by_uuid(&map, info.uuid);
+	if (mp)
+		mdfd = open_dev(mp->devnum);
+	else
+		mdfd = -1;
+
 	if (mdfd < 0) {
-		fprintf(stderr, Name ": failed to open %s: %s.\n",
-			chosen_name, strerror(errno));
-		return 2;
-	}
-	/* 5/ Find out if array already exists */
-	if (! mddev_busy(devnum)) {
-	/* 5a/ if it does not */
-	/* - choose a name, from mdadm.conf or 'name' field in array. */
-	/* - create the array */
-	/* - add the device */
-		mdu_array_info_t ainf;
-		mdu_disk_info_t disk;
-		char md[20];
 		struct mdinfo *sra;
+		struct mdinfo dinfo;
 
-		memset(&ainf, 0, sizeof(ainf));
-		ainf.major_version = st->ss->major;
-		ainf.minor_version = st->minor_version;
-		if (ioctl(mdfd, SET_ARRAY_INFO, &ainf) != 0) {
-			fprintf(stderr, Name
-				": SET_ARRAY_INFO failed for %s: %s\b",
+		/* Couldn't find an existing array, maybe make a new one */
+		mdfd = create_mddev(match ? match->devname : NULL,
+				    name_to_use, autof, trustworthy, chosen_name);
+
+		if (mdfd < 0)
+			return 1;
+
+		sysfs_init(&info, mdfd, 0);
+
+		if (set_array_info(mdfd, st, &info) != 0) {
+			fprintf(stderr, Name ": failed to set array info for %s: %s\n",
 				chosen_name, strerror(errno));
 			close(mdfd);
 			return 2;
 		}
-		sprintf(md, "%d.%d\n", st->ss->major, st->minor_version);
-		sra = sysfs_read(mdfd, devnum, GET_VERSION);
-		sysfs_set_str(sra, NULL, "metadata_version", md);
-		memset(&disk, 0, sizeof(disk));
-		disk.major = major(stb.st_rdev);
-		disk.minor = minor(stb.st_rdev);
-		sysfs_free(sra);
-		if (ioctl(mdfd, ADD_NEW_DISK, &disk) != 0) {
+
+		dinfo = info;
+		dinfo.disk.major = major(stb.st_rdev);
+		dinfo.disk.minor = minor(stb.st_rdev);
+		if (add_disk(mdfd, st, &info, &dinfo) != 0) {
 			fprintf(stderr, Name ": failed to add %s to %s: %s.\n",
 				devname, chosen_name, strerror(errno));
 			ioctl(mdfd, STOP_ARRAY, 0);
 			close(mdfd);
 			return 2;
 		}
-		sra = sysfs_read(mdfd, devnum, GET_DEVS);
+		sra = sysfs_read(mdfd, fd2devnum(mdfd), GET_DEVS);
 		if (!sra || !sra->devs || sra->devs->disk.raid_disk >= 0) {
 			/* It really should be 'none' - must be old buggy
 			 * kernel, and mdadm -I may not be able to complete.
@@ -326,6 +336,12 @@ int Incremental(char *devname, int verbose, int runstop,
 			sysfs_free(sra);
 			return 2;
 		}
+		info.array.working_disks = 1;
+		sysfs_free(sra);
+		/* 6/ Make sure /var/run/mdadm.map contains this array. */
+		map_update(&map, fd2devnum(mdfd),
+			   info.text_version,
+			   info.uuid, chosen_name);
 	} else {
 	/* 5b/ if it does */
 	/* - check one drive in array to make sure metadata is a reasonably */
@@ -333,60 +349,60 @@ int Incremental(char *devname, int verbose, int runstop,
 	/* - add the device */
 		char dn[20];
 		int dfd2;
-		mdu_disk_info_t disk;
 		int err;
 		struct mdinfo *sra;
 		struct supertype *st2;
-		sra = sysfs_read(mdfd, devnum, (GET_VERSION | GET_DEVS |
-						GET_STATE));
+		struct mdinfo info2, *d;
 
-		if (sra->array.major_version != st->ss->major ||
-		    sra->array.minor_version != st->minor_version) {
-			if (verbose >= 0)
+		if (mp->path)
+			strcpy(chosen_name, mp->path);
+		else
+			strcpy(chosen_name, devnum2devname(mp->devnum));
+
+		sra = sysfs_read(mdfd, fd2devnum(mdfd), (GET_DEVS | GET_STATE));
+
+		if (sra->devs) {
+			sprintf(dn, "%d:%d", sra->devs->disk.major,
+				sra->devs->disk.minor);
+			dfd2 = dev_open(dn, O_RDONLY);
+			st2 = dup_super(st);
+			if (st2->ss->load_super(st2, dfd2, NULL) ||
+			    st->ss->compare_super(st, st2) != 0) {
 				fprintf(stderr, Name
-	      ": %s has different metadata to chosen array %s %d.%d %d.%d.\n",
-					devname, chosen_name,
-					sra->array.major_version,
-					sra->array.minor_version,
-					st->ss->major, st->minor_version);
-			close(mdfd);
-			return 1;
-		}
-		sprintf(dn, "%d:%d", sra->devs->disk.major,
-			sra->devs->disk.minor);
-		dfd2 = dev_open(dn, O_RDONLY);
-		st2 = dup_super(st);
-		if (st2->ss->load_super(st2, dfd2, NULL)) {
-			fprintf(stderr, Name
-				": Strange error loading metadata for %s.\n",
-				chosen_name);
-			close(mdfd);
+					": metadata mismatch between %s and "
+					"chosen array %s\n",
+					devname, chosen_name);
+				close(mdfd);
+				close(dfd2);
+				return 2;
+			}
 			close(dfd2);
-			return 2;
+			memset(&info2, 0, sizeof(info2));
+			st2->ss->getinfo_super(st2, &info2);
+			st2->ss->free_super(st2);
+			if (info.array.level != info2.array.level ||
+			    memcmp(info.uuid, info2.uuid, 16) != 0 ||
+			    info.array.raid_disks != info2.array.raid_disks) {
+				fprintf(stderr, Name
+					": unexpected difference between %s and %s.\n",
+					chosen_name, devname);
+				close(mdfd);
+				return 2;
+			}
 		}
-		close(dfd2);
-		st2->ss->getinfo_super(st2, &info2);
-		st2->ss->free_super(st2);
-		if (info.array.level != info2.array.level ||
-		    memcmp(info.uuid, info2.uuid, 16) != 0 ||
-		    info.array.raid_disks != info2.array.raid_disks) {
-			fprintf(stderr, Name
-				": unexpected difference between %s and %s.\n",
-				chosen_name, devname);
-			close(mdfd);
-			return 2;
-		}
-		memset(&disk, 0, sizeof(disk));
-		disk.major = major(stb.st_rdev);
-		disk.minor = minor(stb.st_rdev);
-		err = ioctl(mdfd, ADD_NEW_DISK, &disk);
+		info2.disk.major = major(stb.st_rdev);
+		info2.disk.minor = minor(stb.st_rdev);
+		/* add disk needs to know about containers */
+		if (st->ss->external)
+			sra->array.level = LEVEL_CONTAINER;
+		err = add_disk(mdfd, st, sra, &info2);
 		if (err < 0 && errno == EBUSY) {
 			/* could be another device present with the same
 			 * disk.number. Find and reject any such
 			 */
 			find_reject(mdfd, st, sra, info.disk.number,
 				    info.events, verbose, chosen_name);
-			err = ioctl(mdfd, ADD_NEW_DISK, &disk);
+			err = add_disk(mdfd, st, sra, &info2);
 		}
 		if (err < 0) {
 			fprintf(stderr, Name ": failed to add %s to %s: %s.\n",
@@ -394,25 +410,47 @@ int Incremental(char *devname, int verbose, int runstop,
 			close(mdfd);
 			return 2;
 		}
+		info.array.working_disks = 0;
+		for (d = sra->devs; d; d=d->next)
+			info.array.working_disks ++;
+			
 	}
-	/* 6/ Make sure /var/run/mdadm.map contains this array. */
-	map_update(&map, devnum,
-		   info.array.major_version,
-		   info.array.minor_version,
-		   info.uuid, chosen_name);
 
 	/* 7/ Is there enough devices to possibly start the array? */
 	/* 7a/ if not, finish with success. */
+	if (info.array.level == LEVEL_CONTAINER) {
+		/* Try to assemble within the container */
+		map_unlock(&map);
+		sysfs_uevent(&info, "change");
+		if (verbose >= 0)
+			fprintf(stderr, Name
+				": container %s now has %d devices\n",
+				chosen_name, info.array.working_disks);
+		wait_for(chosen_name, mdfd);
+		close(mdfd);
+		if (runstop < 0)
+			return 0; /* don't try to assemble */
+		rv = Incremental(chosen_name, verbose, runstop,
+				 NULL, homehost, require_homehost, autof);
+		if (rv == 1)
+			/* Don't fail the whole -I if a subarray didn't
+			 * have enough devices to start yet
+			 */
+			rv = 0;
+		return rv;
+	}
 	avail = NULL;
 	active_disks = count_active(st, mdfd, &avail, &info);
 	if (enough(info.array.level, info.array.raid_disks,
 		   info.array.layout, info.array.state & 1,
-		   avail, active_disks) == 0) {
+		   avail, active_disks) == 0 ||
+	    (runstop < 0 && active_disks < info.array.raid_disks)) {
 		free(avail);
 		if (verbose >= 0)
 			fprintf(stderr, Name
 			     ": %s attached to %s, not enough to start (%d).\n",
 				devname, chosen_name, active_disks);
+		map_unlock(&map);
 		close(mdfd);
 		return 0;
 	}
@@ -423,18 +461,18 @@ int Incremental(char *devname, int verbose, int runstop,
 	/*             are enough, */
 	/*   + add any bitmap file  */
 	/*   + start the array (auto-readonly). */
-{
-	mdu_array_info_t ainf;
 
 	if (ioctl(mdfd, GET_ARRAY_INFO, &ainf) == 0) {
 		if (verbose >= 0)
 			fprintf(stderr, Name
 			   ": %s attached to %s which is already active.\n",
 				devname, chosen_name);
-		close (mdfd);
+		close(mdfd);
+		map_unlock(&map);
 		return 0;
 	}
-}
+
+	map_unlock(&map);
 	if (runstop > 0 || active_disks >= info.array.working_disks) {
 		struct mdinfo *sra;
 		/* Let's try to start it */
@@ -457,9 +495,9 @@ int Incremental(char *devname, int verbose, int runstop,
 			}
 			close(bmfd);
 		}
-		sra = sysfs_read(mdfd, devnum, 0);
+		sra = sysfs_read(mdfd, fd2devnum(mdfd), 0);
 		if ((sra == NULL || active_disks >= info.array.working_disks)
-		    && name != NULL)
+		    && trustworthy != FOREIGN)
 			rv = ioctl(mdfd, RUN_ARRAY, NULL);
 		else
 			rv = sysfs_set_str(sra, NULL,
@@ -470,6 +508,7 @@ int Incremental(char *devname, int verbose, int runstop,
 			   ": %s attached to %s, which has been started.\n",
 					devname, chosen_name);
 			rv = 0;
+			wait_for(chosen_name, mdfd);
 		} else {
 			fprintf(stderr, Name
                              ": %s attached to %s, but failed to start: %s.\n",
@@ -603,53 +642,6 @@ static int count_active(struct supertype *st, int mdfd, char **availp,
 	return cnt + cnt1;
 }
 
-void RebuildMap(void)
-{
-	struct mdstat_ent *mdstat = mdstat_read(0, 0);
-	struct mdstat_ent *md;
-	struct map_ent *map = NULL;
-	int mdp = get_mdp_major();
-
-	for (md = mdstat ; md ; md = md->next) {
-		struct mdinfo *sra = sysfs_read(-1, md->devnum, GET_DEVS);
-		struct mdinfo *sd;
-
-		for (sd = sra->devs ; sd ; sd = sd->next) {
-			char dn[30];
-			int dfd;
-			int ok;
-			struct supertype *st;
-			char *path;
-			struct mdinfo info;
-
-			sprintf(dn, "%d:%d", sd->disk.major, sd->disk.minor);
-			dfd = dev_open(dn, O_RDONLY);
-			if (dfd < 0)
-				continue;
-			st = guess_super(dfd);
-			if ( st == NULL)
-				ok = -1;
-			else
-				ok = st->ss->load_super(st, dfd, NULL);
-			close(dfd);
-			if (ok != 0)
-				continue;
-			st->ss->getinfo_super(st, &info);
-			if (md->devnum > 0)
-				path = map_dev(MD_MAJOR, md->devnum, 0);
-			else
-				path = map_dev(mdp, (-1-md->devnum)<< 6, 0);
-			map_add(&map, md->devnum, st->ss->major,
-				st->minor_version,
-				info.uuid, path ? : "/unknown");
-			st->ss->free_super(st);
-			break;
-		}
-	}
-	map_write(map);
-	map_free(map);
-}
-
 int IncrementalScan(int verbose)
 {
 	/* look at every device listed in the 'map' file.
@@ -667,12 +659,11 @@ int IncrementalScan(int verbose)
 	devs = conf_get_ident(NULL);
 
 	for (me = mapl ; me ; me = me->next) {
-		char path[1024];
 		mdu_array_info_t array;
 		mdu_bitmap_file_t bmf;
 		struct mdinfo *sra;
-		int mdfd = open_mddev_devnum(me->path, me->devnum,
-					     NULL, path, 0);
+		int mdfd = open_dev(me->devnum);
+
 		if (mdfd < 0)
 			continue;
 		if (ioctl(mdfd, GET_ARRAY_INFO, &array) == 0 ||
@@ -682,7 +673,8 @@ int IncrementalScan(int verbose)
 		}
 		/* Ok, we can try this one.   Maybe it needs a bitmap */
 		for (mddev = devs ; mddev ; mddev = mddev->next)
-			if (strcmp(mddev->devname, me->path) == 0)
+			if (mddev->devname && me->path
+			    && devname_matches(mddev->devname, me->path))
 				break;
 		if (mddev && mddev->bitmap_file) {
 			/*
@@ -716,14 +708,138 @@ int IncrementalScan(int verbose)
 				if (verbose >= 0)
 					fprintf(stderr, Name
 						": started array %s\n",
-						me->path);
+						me->path ?: devnum2devname(me->devnum));
 			} else {
 				fprintf(stderr, Name
 					": failed to start array %s: %s\n",
-					me->path, strerror(errno));
+					me->path ?: devnum2devname(me->devnum),
+					strerror(errno));
 				rv = 1;
 			}
 		}
 	}
 	return rv;
 }
+
+static char *container2devname(char *devname)
+{
+	char *mdname = NULL;
+
+	if (devname[0] == '/') {
+		int fd = open(devname, O_RDONLY);
+		if (fd >= 0) {
+			mdname = devnum2devname(fd2devnum(fd));
+			close(fd);
+		}
+	} else {
+		int uuid[4];
+		struct map_ent *mp, *map = NULL;
+					
+		if (!parse_uuid(devname, uuid))
+			return mdname;
+		mp = map_by_uuid(&map, uuid);
+		if (mp)
+			mdname = devnum2devname(mp->devnum);
+		map_free(map);
+	}
+
+	return mdname;
+}
+
+int Incremental_container(struct supertype *st, char *devname, int verbose,
+			  int runstop, int autof, int trustworthy)
+{
+	/* Collect the contents of this container and for each
+	 * array, choose a device name and assemble the array.
+	 */
+
+	struct mdinfo *list = st->ss->container_content(st);
+	struct mdinfo *ra;
+	struct map_ent *map = NULL;
+
+	map_lock(&map);
+
+	for (ra = list ; ra ; ra = ra->next) {
+		int mdfd;
+		char chosen_name[1024];
+		struct map_ent *mp;
+		struct mddev_ident_s *match = NULL;
+
+		mp = map_by_uuid(&map, ra->uuid);
+
+		if (mp) {
+			mdfd = open_dev(mp->devnum);
+			if (mp->path)
+				strcpy(chosen_name, mp->path);
+			else
+				strcpy(chosen_name, devnum2devname(mp->devnum));
+		} else {
+
+			/* Check in mdadm.conf for container == devname and
+			 * member == ra->text_version after second slash.
+			 */
+			char *sub = strchr(ra->text_version+1, '/');
+			struct mddev_ident_s *array_list;
+			if (sub) {
+				sub++;
+				array_list = conf_get_ident(NULL);
+			} else
+				array_list = NULL;
+			for(; array_list ; array_list = array_list->next) {
+				char *dn;
+				if (array_list->member == NULL ||
+				    array_list->container == NULL)
+					continue;
+				if (strcmp(array_list->member, sub) != 0)
+					continue;
+				if (array_list->uuid_set &&
+				    !same_uuid(ra->uuid, array_list->uuid, st->ss->swapuuid))
+					continue;
+				dn = container2devname(array_list->container);
+				if (dn == NULL)
+					continue;
+				if (strncmp(dn, ra->text_version+1,
+					    strlen(dn)) != 0 ||
+				    ra->text_version[strlen(dn)+1] != '/') {
+					free(dn);
+					continue;
+				}
+				free(dn);
+				/* we have a match */
+				match = array_list;
+				if (verbose>0)
+					fprintf(stderr, Name ": match found for member %s\n",
+						array_list->member);
+				break;
+			}
+
+			if (match && match->devname &&
+			    strcasecmp(match->devname, "<ignore>") == 0) {
+				if (verbose > 0)
+					fprintf(stderr, Name ": array %s/%s is "
+						"explicitly ignored by mdadm.conf\n",
+						match->container, match->member);
+				return 2;
+			}
+			if (match)
+				trustworthy = LOCAL;
+
+			mdfd = create_mddev(match ? match->devname : NULL,
+					    ra->name,
+					    autof,
+					    trustworthy,
+					    chosen_name);
+		}
+
+		if (mdfd < 0) {
+			fprintf(stderr, Name ": failed to open %s: %s.\n",
+				chosen_name, strerror(errno));
+			return 2;
+		}
+
+		assemble_container_content(st, mdfd, ra, runstop,
+					   chosen_name, verbose);
+	}
+	map_unlock(&map);
+	return 0;
+}
